package ParkNPark.common;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;

/**
 * Parses a command line using GNU GetOpt
 */
public class CommandLineParser
{
	/** The reply size of the server messages in this test run in bytes */
	protected int replySize;
	
	/** The number of clients in this test run */
	protected int numClients;
	
	/** The inter-request time in this test run in milliseconds */
	protected int interRequestTime;
	
	/**
	 * Parses the given command line for the given program name and
	 * returns true if the command line was successfully parsed or
	 * false if the program should exit
	 * @param name The name of the program
	 * @param args The arguments passed in from the command line
	 * @return True if the program can continue or false if the program
	 * should exit
	 */
	public boolean parseCommandLine(String name, String[] args)
	{
        LongOpt options[] = new LongOpt[9];
        options[0] = new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h');
        options[1] = new LongOpt("ORBInitialHost", LongOpt.REQUIRED_ARGUMENT, null, 'w');
        options[2] = new LongOpt("ORBInitialPort", LongOpt.REQUIRED_ARGUMENT, null, 'x');
        options[3] = new LongOpt("ORBServerPort", LongOpt.REQUIRED_ARGUMENT, null, 'y');
        options[4] = new LongOpt("ORBServerHost", LongOpt.REQUIRED_ARGUMENT, null, 'z');
        options[5] = new LongOpt("reply-size", LongOpt.REQUIRED_ARGUMENT, null, 'r');
        options[6] = new LongOpt("num-clients", LongOpt.REQUIRED_ARGUMENT, null, 'n');
        options[7] = new LongOpt("interrequest-time", LongOpt.REQUIRED_ARGUMENT, null, 'i');
        options[8] = new LongOpt("test", LongOpt.NO_ARGUMENT, null, 't');
        Getopt opt = new Getopt(name, args, "w:x:y:z:r:n:i:t", options, true);
        int c;
        boolean testingOnly = false;
        while ((c = opt.getopt()) != -1) {
            switch (c)
            {
                case '?' :  // Warning already written out by GetOpt
                	System.out.println();
                case 'h' :
                {
                	// Display the help screen
                	help(name);
                	return false;
                }
                
                case 'r' :
                {
                	// Set the reply size for the test
                	String arg = opt.getOptarg();
                	try {
                	    replySize = Integer.parseInt(arg, 10);
                	    
                	    // Validate the reply size
                	    if (replySize != 0 && replySize != 256 && replySize != 512 &&
                	    	replySize != 1024) {
                	    	System.err.println("Reply size can only be 0 (original), 256, 512, or 1024");
                    		System.err.flush();
                        	System.out.println();
                        	help(name);
                	    	return false;
                	    }
                	} catch (NumberFormatException e) {
                		System.err.println("Parameter reply-size is not a number");
                		System.err.flush();
                    	System.out.println();
                    	help(name);
                		return false;
                	}
                	break;
                }
                
                case 'n' :
                {
                	// Set the number of clients
                	String arg = opt.getOptarg();
                	try {
                    	numClients = Integer.parseInt(arg, 10);
                    	
                    	// Validate the number of clients
                    	if (numClients != 1 && numClients != 4 && numClients != 7 &&
                    		numClients != 10) {
                    		System.err.println("Number of clients can only be 1, 4, 7, or 10");
                    		System.err.flush();
                        	System.out.println();
                        	help(name);
                    		return false;
                    	}
                	} catch (NumberFormatException e) {
                		System.err.println("Parameter num-clients is not a number");
                		System.err.flush();
                    	System.out.println();
                    	help(name);
                		return false;
                	}
                	break;
                }
            
                case 'i' :
                {
                	// Set the inter-request time
                	String arg = opt.getOptarg();
                	try {
                    	interRequestTime = Integer.parseInt(arg, 10);
                    	
                    	// Validate the inter-request time
                    	if (interRequestTime != 0 && interRequestTime != 20 &&
                    		interRequestTime != 40) {
                    		System.err.println("Inter-request time must be 0 (no delay), 20, or 40 milliseconds");
                    		System.err.flush();
                        	System.out.println();
                        	help(name);
                    		return false;
                    	}
                	} catch (NumberFormatException e) {
                		System.err.println("Parameter interrequest-time is not a number");
                		System.err.flush();
                    	System.out.println();
                    	help(name);
                		return false;
                	}
                	break;
                }
                
                case 't' :
                {
                	// The caller merely wanted to test whether or not the command
                	// line is valid or not. We will return success or failure before
                	// this method returns
                	testingOnly = true;
                	break;
                }
            }
        }
        
        // Ensure that numClients is set
        if (numClients == 0) {
        	System.err.println("--num-clients is a required parameter");
    		System.err.flush();
        	System.out.println();
        	help(name);
        	return false;
        }
        
        // If the caller was only testing us, then exit successfully now
        if (testingOnly) {
        	System.exit(0);
        	return false;
        }
        
        // We can continue
		return true;
	}
	
	/**
	 * Writes out our command line help text to stdout using the given program name
	 * @param name The program name to use when displaying the command line help
	 */
	protected void help(String name) {
    	System.out.println(name + " command line parameters:");
    	System.out.println("");
    	System.out.println("-ORBInitialHost      Host name or IP address of the name service ");
    	System.out.println("-ORBInitialPort      Port of the name service");
    	System.out.println("-ORBServerHost       Host name or IP address to serve on (server only)");
    	System.out.println("-ORBServerPort       Port to serve on (server only)");
    	System.out.println("--reply-size         Size of the reply messages: 0 (original), 256, 512, or");
    	System.out.println("                     1024 bytes");
    	System.out.println("--num-clients        Number of clients running in this test: 1, 4, 7, or 10");
    	System.out.println("--interrequest-time  Time to delay in milliseconds between requests:");
    	System.out.println("                     0 (no delay), 20, or 40");
    	System.out.println();
    	System.out.println("Note that the -ORBServerHost parameter is only necessary if this system's host");
    	System.out.println("name maps to a localhost address such as 127.0.0.1.");
    	System.out.println();
    	System.out.println("The -ORBServerHost and -ORBServerPort parameters are both ignored on clients.");
    	System.out.println("These parameters are only useful when starting a server.");
    	System.out.println();
    	System.out.println("As an example, if we want to start a server with a 256-byte reply size, 7");
    	System.out.println("clients, and an inter-request time of 20 milliseconds, the following command");
    	System.out.println("line can be used (assumes that the name service is on the local host on the");
    	System.out.println("default port):");
    	System.out.println();
    	System.out.println("  java -jar Server.jar --reply-size 256 --num-clients 7 --interrequest-time 20");
    	System.out.println();
    	System.out.println("To start a server on port 900 that communicates with the name service on host");
    	System.out.println("nameservice on port 833 and has a test case of 4 clients, a no-delay");
    	System.out.println("inter-request time, and a reply size of 1024 bytes, use the following command");
    	System.out.println("line:");
    	System.out.println();
    	System.out.println("  java -jar Server.jar -ORBInitialHost nameservice -ORBInitialPort 833 \\");
    	System.out.println("       -ORBServerPort 900 --num-clients 4 --reply-size 1024");
	}

	/**
	 * Returns the inter-request time in this test run
	 * @return The inter-request time in this test run
	 */
	public int getInterRequestTime() {
		return interRequestTime;
	}

	/**
	 * Returns the number of clients in this test run
	 * @return The number of clients in this test run
	 */
	public int getNumClients() {
		return numClients;
	}

	/**
	 * Returns the reply size of the server messages in this test run
	 * @return The reply size of the server messages in this test run
	 */
	public int getReplySize() {
		return replySize;
	}
}
