package ParkNPark.common;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Logger class that, as quickly as possible, logs information into a
 * pre-allocated buffer and has facilities for manual and automatic flushing.
 */
public class Logger
{
	/** Our currently-active log */
	protected List<LogEntry> log;
	
	/** Consistent mutex for the currently-active log (because log can get replaced by a new
	 *  instance) */
	protected Object logMutex = new Object();
	
	/** The number of array elements to allocate when creating new Lists */
	protected int allocationSize;
	
	/** Sink for the times that method calls begin in */
	protected BufferedWriter probeIn;
	
	/** Sink for the times that method calls complete in */
	protected BufferedWriter probeOut;
	
	/** Method name sink */
	protected BufferedWriter probeMethod;
	
	/** Client name sink */
	protected BufferedWriter probeSource;
	
	/**
	 * Creates a new Logger instance that can be used to log activities and,
	 * in the background, flush existing log entries to disk
	 * @param numClients The number of clients in the current test
	 * @param requestsPerClient The number of requests expected from each client
	 * @param interRequestTime The delay between each invocation, which is used when
	 * constructing the log file names
	 * @param replySize The size of the server's replies, which is used when
	 * constructing the log file names
	 * @param tier The tier being logged: <code>srv</code> or <code>cli</code>
	 * @param hostname The host name of the machine being logged
	 */
	public Logger(int numClients, int requestsPerClient, int interRequestTime, int replySize,
			      String tier, String hostname, boolean logSource)
	{
		// Store our array allocation size
		allocationSize = numClients * requestsPerClient;
		
		// Create our initial log
		log = new ArrayList<LogEntry>(allocationSize);
		
		// Create our log files
		try {
			probeIn = new BufferedWriter(new FileWriter(generateLogFileName(numClients, interRequestTime, replySize, "in", tier, hostname)));
			probeOut = new BufferedWriter(new FileWriter(generateLogFileName(numClients, interRequestTime, replySize, "out", tier, hostname)));
			probeMethod = new BufferedWriter(new FileWriter(generateLogFileName(numClients, interRequestTime, replySize, "msg", tier, hostname)));
			if (logSource)
			    probeSource = new BufferedWriter(new FileWriter(generateLogFileName(numClients, interRequestTime, replySize, "source", tier, hostname)));
			else
				probeSource = null;
		}
		catch (IOException e) {
			System.err.println("Could not create probe log files; exiting");
			System.err.flush();
			throw new RuntimeException("Could not create probe log files");
		}
	}
	
	/**
	 * Creates a new LogEntry object instance, adds it to the log,
	 * and, just before returning it, sets its probeIn property to the
	 * current time in microseconds. Use this method only if you
	 * constructed this Logger instance with logSource set to false
	 * @param methodName The name of the method that was called
	 * @return A new LogEntry with the
	 * current time in microseconds into the probeIn property
	 */
	public LogEntry beginLogEntry(String methodName) {
		return beginLogEntry(methodName, null);
	}
	
	/**
	 * Creates a new LogEntry object instance, adds it to the log,
	 * and, just before returning it, sets its probeIn property to the
	 * current time in microseconds
	 * @param methodName The name of the method that was called
	 * @param clientName The hostname of the client that called the method
	 * @return A new LogEntry with the
	 * current time in microseconds into the probeIn property
	 */
	public LogEntry beginLogEntry(String methodName, String clientName)
	{
		// Create the new LogEntry instance
		LogEntry logEntry = new LogEntry();
		
		// Serialize access to the log
		boolean flush = false;
		do {
			synchronized(logMutex)
			{
				// If the log is full, warn the user and flush the logs first
				if (allocationSize <= log.size())
					flush = true;
				
				// Otherwise, add it to the log
				else {
				    log.add(logEntry);
				    flush = false;
				}

				// If we need to flush, then do that now
				if (flush) {
					System.err.println("Logger Warning: Log is full; flushing contents before beginning the new log entry");
				    System.err.flush();
					try {
					    flush();
					    System.err.println("Flush complete; resuming");
					    System.err.flush();
					} catch (IOException e) {
						System.err.println("IOException in Logger.flush(): " + e.getMessage());
						System.err.println("Recent log data might be permanently lost");
						System.err.flush();
					}
				}
			}
		} while (flush);

		// Set the method name and the client name
		logEntry.setProbeMethod(methodName);
		logEntry.setProbeSource(clientName);
		
		// Set the probeIn property and immediately return
		logEntry.setProbeIn(System.nanoTime() / 1000);
		return logEntry;
	}
	
	/**
	 * Ends a LogEntry after setting the probeOut property to the current
	 * time in microseconds
	 * @param logEntry The LogEntry instance to finish
	 */
	public void endLogEntry(LogEntry logEntry)
	{
		// Set the probeOut property immediately
		logEntry.setProbeOut(System.nanoTime() / 1000);
		
		// We are finished with this log entry
		logEntry.markLogEntryAsFinished();
	}
	
	/**
	 * Flush out all the log entries into the output files. Do not call this
	 * method if you have <code>logMutex</code> locked or else a deadlock might occur!
	 * @throws IOException Thrown if an IOException occurs during the flushing,
	 * such as out-of-disk-space exceptions
	 */
	public synchronized void flush() throws IOException
	{
		// Perform a synchronized replace of the log
		List<LogEntry> oldLog;
		synchronized(logMutex) {
			List<LogEntry> newLog = new ArrayList<LogEntry>(allocationSize);
			oldLog = log;
			log = newLog;
		}
		
		// Now that we have exclusive use of the old log as no one else has
		// a reference to it, iterate through its entire contents
		for (LogEntry entry : oldLog)
		{
			// If this log entry is not ready, then punt it back to the live log
			if (!entry.isLogEntryFinished()) {
				synchronized(logMutex) {
					log.add(entry);
				}
			}
			
			// Otherwise, write the contents of this log entry to the proper files
			else {
				probeIn.write(Long.toString(entry.getProbeIn(), 10));
				probeIn.newLine();
				probeOut.write(Long.toString(entry.getProbeOut(), 10));
				probeOut.newLine();
				probeMethod.write(entry.getProbeMethod());
				probeMethod.newLine();
				
				// If we are tracking client hostnames, then write that
				if (probeSource != null) {
				    probeSource.write(entry.getProbeSource());
					probeSource.newLine();
				}
			}
		}
		
		// Flush the log file buffers
		probeIn.flush();
		probeOut.flush();
		probeMethod.flush();
		if (probeSource != null)
			probeSource.flush();
	}
	
	/**
	 * Closes the log files. Note that you must not call flush() nor log any more
	 * data after calling this method. Attempts to do so will throw NullPointerExceptions
	 */
	public synchronized void close() throws IOException
	{
		// Perform a flush first
 	    flush();
		
		// Close our outputs
		probeIn.close();
		probeOut.close();
		probeMethod.close();
		if (probeSource != null)
			probeSource.close();
		
		// If we still have unfinished entries, then write out their method names to the console
		// so that the console user can see that bugs might exist in the code (or
		// the server is being closed in the middle of a client request)
		synchronized (logMutex) {
			if (log.size() > 0) {
				Set<String> methodNames = new HashSet<String>();
				for (LogEntry entry : log)
					methodNames.add(entry.getProbeMethod());
				
				// Notify the console user
				System.out.println("The following methods have one or more open log entries:");
				for (String methodName : methodNames) {
					System.out.print("  ");
					System.out.println(methodName);
				}
			}
		}

		// Lose the reference to our log
		log = null;
	}
	
	/**
	 * Generate the name of the log file name according to the given parameters
	 * @param numClients The number of clients in the current test
	 * @param interRequestTime The delay between each invocation
	 * @param replySize The size of the server's replies
	 * @param probeType The type of probe: <code>in</code>, <code>out</code>,
	 * <code>msg</code>, or <code>source</code>
	 * @param machine The machine being logged: <code>srv</code> or <code>cli</code>
	 * @param hostname The host name of the machine being logged
	 */
	protected String generateLogFileName(int numClients, int interRequestTime, int replySize, String probeType, String machine, String hostname)
	{
		return "DATA749_app_" + probeType + "_" + machine + "_WARM_PASSIVE_2srv_" + numClients
			+ "cli_" + interRequestTime + "us_" + replySize + "req_" + hostname + "_team6.txt";
	}
}
