package ParkNPark.tests;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import ParkNPark.common.CommandLineParser;

/**
 * Class that runs one test configuration
 */
public class RunOne implements Runnable
{
	/**
	 * Variables for real-time analysis
	 */
	
	/** Number of clients */
	protected int numClients;
	
	/** Time between requests */
	protected int interRequestTime;
	
	/** Size of reply */
	protected int replySize;

	/** Singleton instance of the ProcessInfoComparator */
	protected static final ProcessInfoComparator<ProcessInfo> processInfoComparator = new ProcessInfoComparator<ProcessInfo>(); 

	/** Our servers */
	protected Set<ProcessInfo> servers = new TreeSet<ProcessInfo>(processInfoComparator);
	
	/** Our clients */
	protected Set<ProcessInfo> clients = new TreeSet<ProcessInfo>(processInfoComparator);
	
	/** Our name and replication server */
	protected ProcessInfo sacredServer = new ProcessInfo();
	
	/** The project root folder */
	protected String projectRoot;
	
	/** A writer for stderr */
	protected BufferedWriter stdErrWriter = new BufferedWriter(new OutputStreamWriter(System.err));
	
	/**
	 * Run the tests as specified on the command line
	 * @param args The command line arguments
	 */
	public static void main(String[] args)
	{
		// Parse the command line
		CommandLineParser clp = new CommandLineParser();
		if (!clp.parseCommandLine(RunOne.class.getName(), args)) {
			System.exit(1);
			return;
		}
		
		// Get our project root from the current working folder
		File projectRoot = new File(System.getProperty("user.dir", ""));
		if (!projectRoot.getName().equals("tests")) {
			System.err.println("Current working folder is not \"tests\"; cannot guarantee directory structure; exiting");
			System.exit(1);
			return;
		}

		// Create our new test and run it
		RunOne test;
		try {
		    test = new RunOne(projectRoot.getParentFile().getAbsolutePath(), clp.getNumClients(), clp.getInterRequestTime(), clp.getReplySize());
		} catch (IOException e) {
			System.err.println(e.getMessage());
			return;
		}
		test.run();
	}
	
	/**
	 * Creates a new test case runner
	 * @param projectRoot The project's root folder to use for this test case run
	 * @param numClients The number of clients to use in this test case run
	 * @param interRequestTime The amount of time in milliseconds to wait between client invocations
	 * @param replySize The size of the replies from the server in bytes
	 * @throws IOException Thrown if an IOException is thrown during the processing
	 */
	public RunOne(String projectRoot, int numClients, int interRequestTime, int replySize) throws IOException {
		this.projectRoot = projectRoot;
		this.numClients = numClients;
		this.interRequestTime = interRequestTime;
		this.replySize = replySize;
		
		// Read our lists from files
	    Set<String> names = new TreeSet<String>();
	    populateSet(names, new FileReader("servers"));
	    for (String name : names) {
	    	ProcessInfo info = new ProcessInfo();
	    	info.name = name;
	    	servers.add(info);
	    }
	    names.clear();
	    populateSet(names, new FileReader("clients"));
	    int i = 0;
	    Iterator<String> nameIterator = names.iterator();
	    while (i < numClients && nameIterator.hasNext()) {
	    	ProcessInfo info = new ProcessInfo();
	    	info.name = nameIterator.next();
	    	clients.add(info);
	    	i++;
	    }
	    names.clear();
	    populateSet(names, new FileReader("sacredserver"));
	    sacredServer.name = names.iterator().next();
    }
	
	/**
	 * Populates a Set from individual lines of a reader
	 * @param set The Set to populate
	 * @param input The Reader to read lines from
	 * @throws IOException Thrown if an IOException arises during processing
	 */
	protected void populateSet(Set<String> set, Reader input) throws IOException
	{
		// Populate the set from the input, one line at a time
		BufferedReader reader = new BufferedReader(input);
		String line;
		while ((line = reader.readLine()) != null) {
			line.trim();
			if (line.length() > 0)
				set.add(line);
		}
	}
	
	/**
	 * Ensures that the output folders exist, which the programs
	 * write stdout or stderr to, depending on the program
	 */
	protected void ensureOutputFoldersExist()
	{
		// Ensure that our output folders exist
		File sacredServerOutput = new File("output/sacredServer");
		File clientsOutput = new File("output/clients");
		File serversOutput = new File("output/servers");
		if (!sacredServerOutput.exists())
			sacredServerOutput.mkdirs();
		if (!clientsOutput.exists())
			clientsOutput.mkdirs();
		if (!serversOutput.exists())
			serversOutput.mkdirs();
	}

	/**
	 * Runs this test's configuration by cleaning out any old
	 * server instances, starting the name service from scratch,
	 * starting the replication manager, starting the servers,
	 * starting the clients, waiting for the clients to exit,
	 * telling the servers to exit, and waiting for the servers
	 * to exit
	 */
	public void run()
	{
		try
		{
			// Ensure that our output folders exist
			ensureOutputFoldersExist();
			
			// Describe this test
			System.out.println("Running test scenario with " + numClients + " clients, an inter-request time of " + interRequestTime + " milliseconds, and a reply size of " + replySize + " bytes");
			
			// Shut down the replication manager
			System.out.println("Stopping the existing replication manager");
			sacredServer.process = Runtime.getRuntime().exec("ssh -x boggle \"" + projectRoot + "/sacredServer/killreplicationmanager\"");
			sacredServer.stdout = new InputCopier(new InputStreamReader(sacredServer.process.getInputStream()),
			                                      new FileWriter("output/sacredServer/" + sacredServer.name));
			sacredServer.stdout.start();
			if (sacredServer.process.waitFor() > 0)
				System.out.println("Could not kill the replication manager (possibly already dead)");
			
			// Shut down the name service
			System.out.println("Stopping the name service and resetting its database");
			sacredServer.process = Runtime.getRuntime().exec("ssh -x boggle killall orbd; rm -rf \"" + projectRoot + "/sacredServer/orb.db\"");
			sacredServer.stdout = new InputCopier(new InputStreamReader(sacredServer.process.getInputStream()),
                                                  new FileWriter("output/sacredServer/" + sacredServer.name));
			sacredServer.stdout.start();
			if (sacredServer.process.waitFor() > 0)
				System.out.println("Could not kill the name service (possily already dead)");
			
			// Start the name service and the replication manager
			System.out.println("Starting the name service and replication manager");
			sacredServer.process = Runtime.getRuntime().exec("ssh -x boggle cd \"" + projectRoot + "/sacredServer\"; ./nameservice; sleep 2; ./replicationmanager" +
					 " --num-clients " + numClients +
					 " --interrequest-time " + interRequestTime +
					 " --reply-size " + replySize);
			sacredServer.stdout = new ServerInputCopier(sacredServer, new InputStreamReader(sacredServer.process.getInputStream()),
                                                        new FileWriter("output/sacredServer/" + sacredServer.name));
			sacredServer.stdout.start();
			sacredServer.stderr = new ErrorCopier(sacredServer, new InputStreamReader(sacredServer.process.getErrorStream()),
                                                  stdErrWriter);
			sacredServer.stderr.start();
			
			// Wait for the replication manager to finish starting
			System.out.println("Waiting for the replication manager to finish starting");
			synchronized(sacredServer.interruptThread) {
				while (!sacredServer.ready && !sacredServer.abnormalExit) {
					try {
					    sacredServer.interruptThread.wait();
					} catch (InterruptedException e) { }
				}

				// If the replication manager is not running, then exit now
				if (sacredServer.abnormalExit) {
					System.out.println("Replication manager on " + sacredServer.name + " did not start up successfully; exiting");
					synchronized(sacredServer.stdout) {
						if (sacredServer.stdout.isAlive())
							sacredServer.stdout.join();
					}
					synchronized(sacredServer.stderr) {
						if (sacredServer.stderr.isAlive())
							sacredServer.stderr.join();
					}
					System.exit(1);
					return;
				}
			}

			// Shut down the existing servers
			System.out.println("Shutting down any existing servers");
			for (ProcessInfo server : servers)
			{
				// Shut down the name service
				System.out.println("  " + server.name);
				server.process = Runtime.getRuntime().exec("ssh -x " + server.name + " killall -9 java");
				server.stdout = new ErrorCopier(server, new InputStreamReader(server.process.getInputStream()),
	                                            stdErrWriter);
				server.stdout.start();
				synchronized(server.stdout) {
					if (server.stdout.isAlive())
						server.stdout.join();
				}
			}
			
			// Start each server
			System.out.println("Starting " + servers.size() + " servers");
			for (ProcessInfo server : servers)
			{
				// Start this server instance and remember it
				System.out.println("  " + server.name);
				Process newServer = Runtime.getRuntime().exec("ssh -x " + server.name +
						" cd \"" + projectRoot + "/server\"; ./server --num-clients " + numClients +
						" --interrequest-time " + interRequestTime +
						" --reply-size " + replySize);
				server.process = newServer;
				
				// Start the output copying
				server.stdout = new ServerInputCopier(server, new InputStreamReader(newServer.getInputStream()),
						                              new FileWriter("output/servers/" + server.name));
				server.stdout.start();
				server.stderr = new ErrorCopier(server, new InputStreamReader(newServer.getErrorStream()),
						                        stdErrWriter);
                server.stderr.start();
			}
			
			// Wait for the servers to finish starting
			System.out.println("Waiting for the servers to finish starting");
			for (ProcessInfo server : servers) {
				System.out.println("  " + server.name);
				
				// Wait for this server to finish starting
				synchronized(server.interruptThread) {
					while (!server.ready && !server.abnormalExit) {
						try {
							server.interruptThread.wait();
						} catch (InterruptedException e) { }
					}

					// If the server is not running, then exit now
					if (server.abnormalExit) {
						System.out.println("Server " + server.name + " did not start up successfully; exiting");
						synchronized(server.stdout) {
							if (server.stdout.isAlive())
								server.stdout.join();
						}
						synchronized(server.stderr) {
							if (server.stderr.isAlive())
								server.stderr.join();
						}
						System.exit(1);
						return;
					}
				}
			}
			
			// Start each client
			System.out.println("Starting " + clients.size() + " clients");
			for (ProcessInfo client : clients)
			{
				// Start this client instance and remember it
				System.out.println("  " + client.name);
				Process newClient = Runtime.getRuntime().exec("ssh -x " + client.name +
						" cd \"" + projectRoot + "/client\"; ./client --num-clients " + numClients +
						" --interrequest-time " + interRequestTime +
						" --reply-size " + replySize + " < ../tests/client-test > /dev/null");
				client.process = newClient;
				
				// Start the output copying
				client.stderr = new ClientInputCopier(client, new InputStreamReader(newClient.getErrorStream()),
						                              new FileWriter("output/clients/" + client.name));
				client.stderr.start();
			}
			
			// Wait for each client to complete
			System.out.println("Waiting for the clients to complete");
			Thread currentThread = Thread.currentThread();
			while (clients.size() > 0)
			{
				// Determine if any clients are active and, if so, sleep
				Iterator<ProcessInfo> clientIterator;
				boolean found = false;
				synchronized(currentThread) {
					clientIterator = clients.iterator();
					while (!found && clientIterator.hasNext()) {
						if (!clientIterator.next().exited)
							found = true;
						else
							clientIterator.remove();
					}
					if (found) {
					    try {
							currentThread.wait();
						} catch (InterruptedException e) { }
					}
				}
			}
			
			// Tell each server to exit
			System.out.println("Waiting for the servers to complete");
			for (ProcessInfo server : servers) {
				System.out.println("  " + server.name);
				
				// Tell this server to exit, which will cause it to flush its logs properly
				server.exitExpected = true;
				BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(server.process.getOutputStream()));
				writer.write("x");
				writer.newLine();
				writer.close();
				
				// Wait for the server to exit
				server.process.waitFor();
				synchronized(server.stdout) {
					if (server.stdout.isAlive())
				        server.stdout.join();
				}
				synchronized(server.stderr) {
					if (server.stderr.isAlive())
					    server.stderr.join();
				}
			}
			
			// Close the sacred server and wait for it to exit
			System.out.println("Closing the connection to the replication manager and name service");
			sacredServer.exitExpected = true;
			sacredServer.process.destroy();
			synchronized(sacredServer.stdout) {
				if (sacredServer.stdout.isAlive())
					sacredServer.stdout.join();
			}
			synchronized(sacredServer.stderr) {
				if (sacredServer.stderr.isAlive())
					sacredServer.stderr.join();
			}
			
			// We are done!
			System.out.println("Test scenario complete");
		} catch (InterruptedException e) {
			System.err.println(e.getMessage());
		} catch (IOException e)  {
			System.err.println(e.getMessage());
		}
	}
}
